#################################################################################################
# Empirical application: reanalysis of Chernozhukov and Hansen (2004)
# Authors: H. Kaido and K. Wuthrich
# Data source: Chernozhukov and Hansen (2004) via Chris Hansen
#################################################################################################

### Packages and preliminaries

library(AER)
library(xtable)

rm(list = ls())

# To replicate the results, change the working directory here.
setwd("/Users/hiroakikaido/Dropbox (Personal)/Under Revision/Decentralization/Codes/final_replication_package")

set.seed(123)

#################################################################################################
# Functions 
#################################################################################################

source("algorithms_final.R")

#################################################################################################
# Data 
#################################################################################################

data.original <- read.table("restatw.dat",header=T)
dat           <- data.original[data.original$inc>=0,]

n       <- nrow(dat)
y       <- dat$net_tfa
d       <- dat$p401
dp      <- d+1
z       <- dat$e401
dat$inc <- dat$inc/10000

# Specification of Chernozhukov and Hansen (2004): 
x <- cbind(dat$i2,dat$i3,dat$i4,dat$i5,dat$i6,dat$i7,dat$a2,dat$a3,dat$a4,dat$a5,dat$fsize,dat$hs,dat$smcol,dat$col,dat$marr,dat$twoearn,dat$db,dat$pira,dat$hown)
k <- dim(x)[2]+1

#################################################################################################
# Analysis 
#################################################################################################

### Specification

taus    <- seq(0.15,0.85,0.05)
bsreps  <- 500
box     <- 30000

### Starting values

tsls        <- ivreg.fit(cbind(dp,cbind(1,x)),y,cbind(z,cbind(1,x)))  
start.alpha <- tsls$coefficients[1]
start.beta  <- tsls$coefficients[2:(k+1)]

### Graphical analysis fixed point

# Computation

grideval <- seq(0,15000,500)
values25 <- sapply(grideval,M,y=y,d=dp,z=z,x=x,tau=0.25)
values50 <- sapply(grideval,M,y=y,d=dp,z=z,x=x,tau=0.5)
values75 <- sapply(grideval,M,y=y,d=dp,z=z,x=x,tau=0.75)

# Graph

graphics.off()

pdf("Figures/fixed_point_graphical.pdf", pointsize=16,width=8.0,height=8.0)
par(mar=c(6,6,6,6))
plot(c(0,15000),c(0,15000), type="n", ylab=expression(hat(M)(theta[2])), xlab=expression(theta[2]), main="")
lines(seq(-5000,20000,500), seq(-5000,20000,500), lty=1,lwd=2)
lines(grideval, values25, lty=1,lwd=2,col="red")
lines(grideval, values50, lty=1,lwd=2,col="blue")
lines(grideval, values75, lty=1,lwd=2,col="green")
points(c(3749,5756,13057),c(3749,5756,13057),col="black",pch=20)
legend("topleft", c("45 degree line", expression(paste(tau," = 0.25")),expression(paste(tau," = 0.50")),expression(paste(tau," = 0.75"))), lty = c(1,1,1,1), lwd = c(2,2,2,2), col =c("black","red","blue","green"), bty = "n")

dev.off()

### Comparison different approaches

# Computation

proc.iqr    <- NULL
proc.br.seq <- NULL
proc.brent  <- NULL

for (t in 1:length(taus)){
  proc.iqr    <- cbind(proc.iqr,iqr(y,d,z,x,taus[t],seq(start.alpha-box,start.alpha+box,length=500))$alpha)
  proc.br.seq <- cbind(proc.br.seq,br.seq(y,dp,z,x,taus[t],start.alpha)$alpha)
  proc.brent  <- cbind(proc.brent,brent.uniroot(start.alpha-box,start.alpha+box,y,dp,z,x,taus[t])$alpha)
} 

# Graph

graphics.off()

pdf("Figures/qte_401k.pdf", pointsize=15,width=8.0,height=6.0)
plot(range(taus),c(0,25000), type="n", ylab="Net Financial Assets (USD)", xlab="Quantile index", main="")
lines(taus, proc.br.seq, lty=1, lwd=3)
lines(taus, proc.brent, lty=2, lwd=3)
lines(taus, proc.iqr, lty=3, lwd=3)
legend("topleft", c("Contraction",  "Brent", "IQR"), lty = c(1,2,3), lwd = c(3,3,3), bty = "n")
dev.off()

### Bootstrapped confidence bands

# Bootstrap algorithm

brent.bs  <- matrix(NA,bsreps,length(taus)) 
seq.bs    <- matrix(NA,bsreps,length(taus)) 

for (rep in 1:bsreps){
  
  print(rep)
  
  dats  <- dat[sample(n,n,replace=TRUE), ]
  ys    <- dats$net_tfa
  ds    <- dats$p401
  dps   <- ds+1
  zs    <- dats$e401
  xs    <- cbind(dats$i2,dats$i3,dats$i4,dats$i5,dats$i6,dats$i7,dats$a2,dats$a3,dats$a4,dats$a5,dats$fsize,dats$hs,dats$smcol,dats$col,dats$marr,dats$twoearn,dats$db,dats$pira,dats$hown)
  
  tsls.s <- ivreg.fit(cbind(dps,cbind(1,xs)),ys,cbind(zs,cbind(1,xs)))  
  start.alpha.s <- tsls.s$coefficients[1]

  proc.br.seq.s <- NULL 
  proc.brent.s  <- NULL
  
  for (t in 1:length(taus)){
    proc.br.seq.s <- cbind(proc.br.seq.s,br.seq(ys,dps,zs,xs,taus[t],start.alpha.s)$alpha)
    proc.brent.s  <- cbind(proc.brent.s,brent.uniroot(start.alpha.s-box,start.alpha.s+box,ys,dps,zs,xs,taus[t])$alpha)
  }
  
  brent.bs[rep,]  <- proc.brent.s
  seq.bs[rep,]    <- proc.br.seq.s
  
}

sd.proc.brent <- (apply((brent.bs-t(matrix(proc.brent, length(taus), bsreps, byrow = F)))^2, 2, mean,na.rm=TRUE))^0.5
sd.proc.seq   <- (apply((seq.bs-t(matrix(proc.br.seq, length(taus), bsreps, byrow = F)))^2, 2, mean,na.rm=TRUE))^0.5

ci.brent.u  <- proc.brent + 1.96 * sd.proc.brent
ci.brent.l  <- proc.brent - 1.96 * sd.proc.brent
ci.seq.u    <- proc.br.seq + 1.96 * sd.proc.seq
ci.seq.l    <- proc.br.seq - 1.96 * sd.proc.seq


# Graph

graphics.off()

pdf("Figures/qte_401k_ci.pdf", pointsize=15,width=8.0,height=6.0)
plot(range(taus),c(0,25000), type="n", ylab="Net Financial Assets (USD)", xlab="Quantile index", main="")
lines(taus, proc.br.seq, lty=1, lwd=3)
lines(taus, ci.seq.u, lty=1, lwd=3)
lines(taus, ci.seq.l, lty=1, lwd=3)
lines(taus, proc.brent, lty=2, lwd=3)
lines(taus, ci.brent.u, lty=2, lwd=3)
lines(taus, ci.brent.l, lty=2, lwd=3)
legend("topleft", c("Contraction","Brent"), lty = c(1,2),lwd =c(3,3), bty = "n")

dev.off()





