###############################################################################
# Estimation algorithms
# Authors: H. Kaido and K. Wuthrich
#
# Notation: 
#  - d, d1, d2, d3 are endogenous regressors
#  - z, z1, z2, z3 are instruments
#  - x are exogenous regressors
#  - alphas (betas) denote for coeffs on endogenous (exogenous) regressors
###############################################################################

# Required packages

library(quantreg)
library(optimization)
library(cmna)

###############################################################################
# One endogenous variable
###############################################################################

M <- function(a0,y,d,z,x,tau){
  b1      <- rq((y-d*a0)~x,tau=tau)$coeff
  a1      <- rq((y-cbind(1,x)%*%b1)~d-1,tau=tau,weights=z/d)$coeff 
  result  <- a1
  return(result)
}

br.seq <- function(y,d,z,x,tau,alpha){
  tolerance <- .Machine$double.eps^0.5
  test      <- tolerance+1
  beta      <- NULL
  nrep      <- 1
  while((test>tolerance)&(nrep<500)){
    alpha_old <- alpha
    beta      <- rq((y-d*alpha)~x,tau=tau)$coeff
    alpha     <- rq((y-cbind(1,x)%*%beta)~d-1,tau=tau,weights=z/d)$coeff 
    test      <- sqrt((alpha - alpha_old)^2)
    nrep      <- nrep + 1
  }
  ifelse(nrep<500,return(list(alpha = alpha, beta = beta)),return(list(alpha = NA, beta = NA)))
}


iqr <- function(y,d,z,x,tau,grid){
  x   <- cbind(1,x)
  obj <- cbind(grid,matrix(Inf,length(grid),dim(cbind(x))[2]+1))
  for(i in 1:length(grid)){
    obj[i,2:dim(obj)[2]]  <- rq.fit.br(cbind(z,x),y-d*obj[i,1],tau)[[1]]
    obj[i,2]              <- obj[i,2]^2
  }
  results <- obj[obj[,2]==min(obj[,2]),-2]
  coeff   <- results
  return(list(alpha=coeff[1],beta=coeff[-1],obj=obj))
}

brent.uniroot <- function(lower,upper,y,d,z,x,tau){
  
  fnr <- function(a0,y,d,z,x,tau){
    b1      <- rq((y-d*a0)~x,tau=tau)$coeff
    a1      <- rq((y-cbind(1,x)%*%b1)~d-1,tau=tau,weights=z/d)$coeff 
    result  <- a1-a0
    return(result)
  }
  
  if (sign(fnr(lower,y,d,z,x,tau))*sign(fnr(upper,y,d,z,x,tau))<=0){
    alpha <- uniroot(fnr,lower=lower,upper=upper,y=y,d=d,z=z,x=x,tau=tau)$root
    beta  <- rq((y-d*alpha)~x,tau=tau)$coeff
  } else {
    alpha <- NA
    beta  <- NA
  }
  return(list(alpha=alpha,beta=beta))
}

brent.obj <- function(lower,upper,y,d,z,x,tau){
  
  fnr <- function(a,y,d,z,x,tau){
    beta  <- rq((y-d*a)~x,tau=tau,method="br")$coeff
    f     <- mean(((y<=cbind(1,x)%*%beta+d*a)-tau)*z)
    return(f)
  }
  
  if (sign(fnr(lower,y,d,z,x,tau))*sign(fnr(upper,y,d,z,x,tau))<=0){
    alpha   <- uniroot(fnr,lower=lower,upper=upper,y=y,d=d,z=z,x=x,tau=tau)$root
    beta    <- rq((y-d*alpha)~x,tau=tau,method="br")$coeff
  } else {
    alpha <- NA
    beta  <- NA
  }
  
  return(list(alpha=alpha,beta=beta))
  
}

bisection.obj <- function(lower,upper,y,d,z,x,tau){
  
  fnr <- function(alpha){
    beta  <- rq((y-d*alpha)~x,tau=tau,method="br")$coeff
    f     <- mean(((y<=cbind(1,x)%*%beta+d*alpha)-tau)*z)
    return(f)
  }
  
  if (sign(fnr(lower))*sign(fnr(upper))<=0){
    alpha <- bisection(fnr, lower , upper, tol = .Machine$double.eps^0.5, m = 100)
    beta  <- rq((y-d*alpha)~x,tau=tau,method="br")$coeff
  } else {
    alpha <- NA
    beta  <- NA
  }
  
  return(list(alpha=alpha,beta=beta))
  
}

evaluate1 <- function(y,d,z,x,tau,start.alpha,lower,upper,leng){
  
  # Contraction algorithms
  r.br.seq <- br.seq(y,d,z,x,tau,start.alpha)

  # Root-finding algorithms
  r.brent <- brent.uniroot(lower,upper,y,d,z,x,tau)
  
  # Concentration-based approaches
  r.brent.obj     <- brent.obj(lower,upper,y,d,z,x,tau)
  r.bisection.obj <- bisection.obj(lower,upper,y,d,z,x,tau)
 
  # IQR
  r.iqr <- iqr(y,d,z,x,tau,seq(lower,upper,length=leng))

  return(c(r.br.seq$alpha,r.brent$alpha,r.brent.obj$alpha,r.bisection.obj$alpha,r.iqr$alpha))
  
}

###############################################################################
# Two endogenous variables
###############################################################################

br.seq.2 <- function(y,d1,d2,z1,z2,x,tau,alpha1,alpha2){
  tolerance <- .Machine$double.eps^0.5
  test      <- tolerance+1
  beta      <- NULL
  nrep      <- 1
  while((test>tolerance)&(nrep<500)){
    alpha1_old  <- alpha1
    alpha2_old  <- alpha2
    beta        <- rq((y-d1*alpha1-d2*alpha2)~x,tau=tau)$coeff
    alpha1      <- rq((y-d2*alpha2-cbind(1,x)%*%beta)~d1-1,tau=tau,weights=z1/d1)$coeff 
    alpha2      <- rq((y-d1*alpha1-cbind(1,x)%*%beta)~d2-1,tau=tau,weights=z2/d2)$coeff 
   
    test        <- sqrt((alpha1-alpha1_old)^2+(alpha2-alpha2_old)^2)
    nrep        <- nrep + 1
  }
  ifelse(nrep<500,return(list(alpha1 = alpha1, alpha2 = alpha2, beta = beta)),return(list(alpha1 = NA, alpha2 = NA, beta = NA)))
}

sa.2 <- function(y,d1,d2,z1,z2,x,tau,alpha1.start,alpha2.start,lower,upper){
  fn2 <- function(alpha){
    alpha1    <- alpha[1]
    alpha2    <- alpha[2]
    beta.u    <- rq((y-d1*alpha1 - d2*alpha2)~x,tau=tau)$coeff
    alpha2.u  <- rq((y-d1*alpha1 - cbind(1,x)%*%beta.u) ~ d2-1,tau=tau,weights=z2/d2)$coeff
    alpha1.u  <- rq((y-d2*alpha2.u - cbind(1,x)%*%beta.u) ~ d1-1,tau=tau,weights=z1/d1)$coeff
    result    <- sqrt((alpha1.u-alpha1)^2+(alpha2.u-alpha2)^2)
    return(result)
  }
  start.alpha   <- c(alpha1.start,alpha2.start)
  result        <- optim_sa(fn2,start.alpha,lower=lower,upper=upper)
  alpha.result  <- c(result$par)
  beta          <- rq((y-d1*alpha.result[1]-d2*alpha.result[2])~x,tau=tau)$coeff
  return(list(alpha1=alpha.result[1],alpha2=alpha.result[2],beta=beta))
}

brent.uniroot.2 <- function(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2, y, d1, d2, z1, z2, x, tau){
  
  fnr.2 <- function(a1.0,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau){
    r.b     <- brent.uniroot(lower.alpha2,upper.alpha2,y-d1*a1.0,d2,z2,x,tau)
    a1.1    <- rq((y-cbind(1,x)%*%r.b$beta-d2*r.b$alpha)~d1-1,tau=tau,weights=z1/d1)$coeff 
    result  <- a1.1-a1.0
    return(result)
  }
  
  if (sign(fnr.2(lower.alpha1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau))*sign(fnr.2(upper.alpha1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau))<=0){
    alpha1  <- uniroot(fnr.2,lower=lower.alpha1, upper=upper.alpha1,lower.alpha2=lower.alpha2,upper.alpha2=upper.alpha2,y=y,d1=d1,d2=d2,z1=z1,z2=z2,x=x,tau=tau)$root
    r.b     <- brent.uniroot(lower.alpha2,upper.alpha2,y-d1*alpha1, d2, z2, x, tau)
    alpha1  <- alpha1
    alpha2  <- r.b$alpha
    beta    <- r.b$beta
  } else {
    alpha1  <- NA
    alpha2  <- NA
    beta    <- NA
  }
  
  return(list(alpha1=alpha1,alpha2=alpha2,beta=beta))
}


brent.obj.2 <- function(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau){
  
  fnr.2 <- function(a1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau){
    r.b <- brent.obj(lower.alpha2,upper.alpha2,y-d1*a1,d2,z2,x,tau)
    f   <- mean(((y<=cbind(1,x)%*%r.b$beta+d1*a1+d2*r.b$alpha)-tau)*z1)
    return(f)
  } 
  
  if (sign(fnr.2(lower.alpha1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau))*sign(fnr.2(upper.alpha1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau))<=0){
    alpha1  <- uniroot(fnr.2,lower=lower.alpha1, upper=upper.alpha1,lower.alpha2=lower.alpha2,upper.alpha2=upper.alpha2,y=y,d1=d1,d2=d2,z1=z1,z2=z2,x=x,tau=tau)$root
    r.b     <- brent.obj(lower.alpha2,upper.alpha2,y-d1*alpha1, d2, z2, x, tau)
    alpha1  <- alpha1
    alpha2  <- r.b$alpha
    beta    <- r.b$beta
  } else {
    alpha1  <- NA
    alpha2  <- NA
    beta    <- NA
  }
  
  return(list(alpha1=alpha1,alpha2=alpha2,beta=beta))
  
}

bisection.obj.2 <- function(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,y,d1,d2,z1,z2,x,tau){
  
  fnr.2 <- function(a1){
    r.b <- bisection.obj(lower.alpha2,upper.alpha2,y-d1*a1,d2,z2,x,tau)
    f   <- mean(((y<=cbind(1,x)%*%r.b$beta+d1*a1+d2*r.b$alpha)-tau)*z1)
    return(f)
  } 
  
  if (sign(fnr.2(lower.alpha1))*sign(fnr.2(upper.alpha1))<=0){
    alpha1  <- bisection(fnr.2, lower.alpha1 ,upper.alpha1, tol = .Machine$double.eps^0.5, m = 100)
    r.b     <- bisection.obj(lower.alpha2,upper.alpha2,y-d1*alpha1,d2,z2,x,tau)
    alpha2  <- r.b$alpha
    beta    <- r.b$beta
  } else {
    alpha1  <- NA
    alpha2  <- NA
    beta    <- NA
  }
  
  return(list(alpha1=alpha1,alpha2=alpha2,beta=beta))
  
}

iqr.2 <- function(y,d1,d2,z1,z2,x,tau,grid1,grid2){
  grid    <- expand.grid(grid1,grid2)
  k       <- dim(cbind(x))[2]
  results <- matrix(NA,dim(grid)[1],k+4)
  p       <- 1
  for (i in 1:dim(grid)[1]){
    qr.r        <- try(rq(y-cbind(d1,d2)%*%t(grid[p,])~x+z1+z2,tau),TRUE)
    OMEGA       <- summary.rq(qr.r,covariance=T)$cov[(k+2):(k+3),(k+2):(k+3)]
    crit        <- rbind(qr.r$coeff[(k+2):(k+3)])%*%solve(OMEGA)%*%cbind(qr.r$coeff[(k+2):(k+3)])
    results[p,] <- cbind(crit,grid[p,1],grid[p,2],rbind(qr.r$coeff[1:(k+1)]))
    p           <- p+1  
  }
  coeff <- results[results[,1]==min(results[,1]),-1]
  return(list(alpha1=coeff[1], alpha2=coeff[2], beta=coeff[-(1:2)]))
}

evaluate2 <- function(y,d1,d2,z1,z2,x,tau,start.alpha1,start.alpha2,lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,leng){
  
  # Contraction algorithms
  r.br.seq <- br.seq.2(y,d1,d2,z1,z2,x,tau,start.alpha1,start.alpha2)

  # Root-finding algorithms (implemented through optimizers)
  r.sa <- sa.2(y,d1,d2,z1,z2,x,tau,start.alpha1,start.alpha2,c(lower.alpha1,lower.alpha2),c(upper.alpha1,upper.alpha2))
  
  # Nested algorithm
  r.brent <- try(brent.uniroot.2(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2, y, d1, d2, z1, z2, x, tau),TRUE)
  if (is.list(r.brent[1])==TRUE){
    brent.alpha1 <- r.brent$alpha1
    brent.alpha2 <- r.brent$alpha2
  } else {
    brent.alpha1 <- NA
    brent.alpha2 <- NA
  }
  
  # Nested concentrating-out approach
  r.brent.obj <- try(brent.obj.2(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2, y, d1, d2, z1, z2, x, tau),TRUE)
  if (is.list(r.brent.obj[1])==TRUE){
    brent.obj.alpha1 <- r.brent.obj$alpha1
    brent.obj.alpha2 <- r.brent.obj$alpha2
  } else {
    brent.obj.alpha1 <- NA
    brent.obj.alpha2 <- NA
  }
  
  # IQR
  r.iqr.2 <- try(iqr.2(y,d1,d2,z1,z2,x,tau,seq(lower.alpha1,upper.alpha1,length=leng),seq(lower.alpha2,upper.alpha2,length=leng)),TRUE)
  if (is.list(r.iqr.2[1])==TRUE){
    iqr.alpha1 <- r.iqr.2$alpha1
    iqr.alpha2 <- r.iqr.2$alpha2
  } else {
    iqr.alpha1 <- NA
    iqr.alpha2 <- NA
  }
  
  return(c(r.br.seq$alpha1,brent.alpha1,r.sa$alpha1,brent.obj.alpha1,iqr.alpha1,
           r.br.seq$alpha2,brent.alpha2,r.sa$alpha2,brent.obj.alpha2,iqr.alpha2))
  
}

###############################################################################
# Three endogenous variables
###############################################################################

br.seq.3 <- function(y,d1,d2,d3,z1,z2,z3,x,tau,alpha1,alpha2,alpha3){
  tolerance <- .Machine$double.eps^0.5
  test <- tolerance+1
  beta <- NULL
  nrep <- 1
  while((test>tolerance)&(nrep<500)){
    alpha1_old  <- alpha1
    alpha2_old  <- alpha2
    alpha3_old  <- alpha3
    beta        <- rq((y-d1*alpha1-d2*alpha2-d3*alpha3)~x,tau=tau)$coeff
    alpha1      <- rq((y-d3*alpha3-d2*alpha2-cbind(1,x)%*%beta)~d1-1,tau=tau,weights=z1/d1)$coeff 
    alpha2      <- rq((y-d3*alpha3-d1*alpha1-cbind(1,x)%*%beta)~d2-1,tau=tau,weights=z2/d2)$coeff 
    alpha3      <- rq((y-d2*alpha2-d1*alpha1-cbind(1,x)%*%beta)~d3-1,tau=tau,weights=z3/d3)$coeff 
    
    test <- sqrt((alpha1-alpha1_old)^2+(alpha2-alpha2_old)^2+(alpha3-alpha3_old)^2)
    nrep <- nrep + 1
  }
  ifelse(nrep<500,return(list(alpha1=alpha1,alpha2=alpha2,alpha3=alpha3,beta=beta)),return(list(alpha1=NA,alpha2=NA,alpha3=NA,beta=NA)))
}

brent.uniroot.3 <- function(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,y,d1,d2,d3,z1,z2,z3,x,tau){
  
  fnr.3 <- function(a1.0,lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,y,d1,d2,d3,z1,z2,z3,x,tau){
    r.b     <- brent.uniroot.2(lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,y-d1*a1.0,d2,d3,z2,z3,x,tau)
    a1.1    <- rq((y-cbind(1,x)%*%r.b$beta-d2*r.b$alpha1-d3*r.b$alpha2)~d1-1,tau=tau,weights=z1/d1)$coeff 
    result  <- a1.1-a1.0
    return(result)
  }
  
  if (sign(fnr.3(lower.alpha1,lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,y,d1,d2,d3,z1,z2,z3,x,tau))*sign(fnr.3(upper.alpha1,lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,y,d1,d2,d3,z1,z2,z3,x,tau))<=0){
    alpha1  <- uniroot(fnr.3,lower=lower.alpha1,upper=upper.alpha1,lower.alpha2=lower.alpha2,upper.alpha2=upper.alpha2,lower.alpha3=lower.alpha3,upper.alpha3=upper.alpha3,y=y,d1=d1,d2=d2,d3=d3,z1=z1,z2=z2,z3=z3,x=x,tau=tau)$root
    r.b     <- brent.uniroot.2(lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,y-d1*alpha1,d2,d3,z2,z3,x,tau)
    alpha1  <- alpha1
    alpha2  <- r.b$alpha1
    alpha3  <- r.b$alpha2
    beta    <- r.b$beta
  } else {
    alpha1  <- NA
    alpha2  <- NA
    alpha3  <- NA
    beta    <- NA
  }
  
  return(list(alpha1=alpha1,alpha2=alpha2,alpha3=alpha3,beta=beta))
}




