#####################################################################################
# Simulations with three endogenous variables
# DPG: 401(k)-Application in Chernozhukov and Hansen (2004)
# Authors: H. Kaido and K. Wuthrich
# Data source: Chernozhukov and Hansen (2004) via Chris Hansen
#####################################################################################

#####################################################################################
# Packages and preliminaries
#####################################################################################

rm(list = ls())

library(AER)
library(xtable)
library(fitdistrplus)
library(MASS)
library(doParallel)
library(doRNG)
registerDoParallel(cores=4)

# To replicate the results, change the working directory here.
setwd("/Users/kasparwuthrich/Dropbox/research/Decentralization IVQR/code/final_replication_package")

source("algorithms_final.R")

#####################################################################################
# Calibrate application-based simulations
#####################################################################################

# Load data

restatw   <- read.table("restatw.dat",header=T)
data.restat      <- restatw[restatw$inc>=0,]
attach(data.restat)

n   <- nrow(data.restat)
y   <- net_tfa
d   <- p401
dp  <- d+1
z   <- e401
inc <- inc/10000
x   <- cbind(inc,age)
k   <- dim(x)[2]+1

# Estimate the distribution of residuals

iqr50       <- iqr(y,d,z,x,0.5,seq(0,10000,length=500)) # IQR
beta50      <- iqr50$beta 
alpha50     <- iqr50$alpha
uhat        <- y-cbind(1,x)%*%beta50-d*alpha50
minuhat     <- min(uhat)
gammaparams <- fitdist(as.vector(uhat-minuhat+1), "gamma", method="mme") 

#####################################################################################
# Simulation wrapper functions
#####################################################################################

sim3 <- function(x,beta50,gammaparams,minuhat,n,nr,tau,box){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  d1r   <- zr*(0.6*temp<ur) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  d1rp  <- d1r+1
  z1r   <- zr
  z2r   <- rnorm(nr)
  z2rp  <- z2r-min(z2r)+1
  z3r   <- rnorm(nr)
  z3rp  <- z3r-min(z3r)+1
  d2r   <- 0.8*z2r + 0.2*qnorm(ur)
  d2rp  <- d2r-min(d2r)+1
  d3r   <- 0.8*z3r + 0.2*qnorm(ur)
  d3rp  <- d3r-min(d3r)+1
  
  yr <- cbind(1,xr)%*%beta50+d1r*(5000+ur*10000)+d2r*10000+d3r*10000+minuhat-1+qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls <- ivreg.fit(cbind(d1rp,d2rp,d3rp,cbind(1,xr)),yr,cbind(z1r,z2r,z3r,cbind(1,xr)))  
  
  start.alpha1 <- tsls$coefficients[1]
  start.alpha2 <- tsls$coefficients[2]
  start.alpha3 <- tsls$coefficients[3]
  
  lower.alpha1 <- start.alpha1-box
  upper.alpha1 <- start.alpha1+box
  lower.alpha2 <- start.alpha2-box
  upper.alpha2 <- start.alpha2+box
  lower.alpha3 <- start.alpha3-box
  upper.alpha3 <- start.alpha3+box
  
  res.brent <- brent.uniroot.3(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,yr,d1rp,d2rp,d3rp,z1r,z2rp,z3rp,xr,tau)
  res.br.seq <- br.seq.3(yr,d1rp,d2rp,d3rp,z1r,z2rp,z3rp,xr,tau,start.alpha1,start.alpha2,start.alpha3)
  
  return(c(res.br.seq$alpha1,res.brent$alpha1,res.br.seq$alpha2,res.brent$alpha2,res.br.seq$alpha3,res.brent$alpha3))
  
}

sim3.ct <- function(x,beta50,gammaparams,minuhat,n,nr,tau,box){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  d1r   <- zr*(0.6*temp<ur) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  d1rp  <- d1r+1
  z1r   <- zr
  z2r   <- rnorm(nr)
  z2rp  <- z2r-min(z2r)+1
  z3r   <- rnorm(nr)
  z3rp  <- z3r-min(z3r)+1
  d2r   <- 0.8*z2r + 0.2*qnorm(ur)
  d2rp  <- d2r-min(d2r)+1
  d3r   <- 0.8*z3r + 0.2*qnorm(ur)
  d3rp  <- d3r-min(d3r)+1
  
  yr <- cbind(1,xr)%*%beta50+d1r*(5000+ur*10000)+d2r*10000+d3r*10000+minuhat-1+qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls <- ivreg.fit(cbind(d1rp,d2rp,d3rp,cbind(1,xr)),yr,cbind(z1r,z2r,z3r,cbind(1,xr)))  
  
  start.alpha1 <- tsls$coefficients[1]
  start.alpha2 <- tsls$coefficients[2]
  start.alpha3 <- tsls$coefficients[3]
  
  lower.alpha1 <- start.alpha1-box
  upper.alpha1 <- start.alpha1+box
  lower.alpha2 <- start.alpha2-box
  upper.alpha2 <- start.alpha2+box
  lower.alpha3 <- start.alpha3-box
  upper.alpha3 <- start.alpha3+box
  
  t.br.seq <- system.time(br.seq.3(yr,d1rp,d2rp,d3rp,z1r,z2rp,z3rp,xr,tau,start.alpha1,start.alpha2,start.alpha3))[3]
  t.brent <- system.time(brent.uniroot.3(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,lower.alpha3,upper.alpha3,yr,d1rp,d2rp,d3rp,z1r,z2rp,z3rp,xr,tau))[3]

  results <- c(t.br.seq,t.brent)
  
  return(results)
  
}


#####################################################################################
# Simulations
#####################################################################################

### Specifications

set.seed(123) 
nreps <- 500
taus <- c(0.15,0.25,0.5,0.75,0.85)
nr <- n
box <- 30000

### Simulations

start <- Sys.time()

dim.eval3 <- 6

bias.3.mat <- var.3.mat <- rmse.3.mat <- convergence.3.mat <- matrix(NA,length(taus),dim.eval3)

for (j in 1:length(taus)){
  
  tau <- taus[j]
  print(tau)
  
  results.3.par <- foreach(i=1:nreps) %dorng% {
    results.3.par <- sim3(x,beta50,gammaparams,minuhat,n,n,tau,box)
  }
  
  alpha10 <- (5000+tau*10000)  
  alpha20 <- 10000
  alpha30 <- 10000
  
  results.3 <- t(matrix(unlist(results.3.par),dim.eval3,nreps))
  bias.3 <- apply(results.3,2,mean,na.rm=TRUE) - c(rep(alpha10,dim.eval3/3),rep(alpha20,dim.eval3/3),rep(alpha30,dim.eval3/3))
  var.3 <- apply(results.3,2,var,na.rm=TRUE)
  rmse.3 <- sqrt(bias.3^2 + var.3)
  convergence.3 <- colMeans(apply(results.3,2,is.na))
  
  bias.3.mat[j,] <- bias.3
  var.3.mat[j,] <- var.3
  rmse.3.mat[j,] <- rmse.3
  convergence.3.mat[j,] <- convergence.3
  
}  

alpha1.results <- cbind(bias.3.mat[,(1:(dim.eval3/3))]/100,rmse.3.mat[,(1:(dim.eval3/3))]/1000)
alpha2.results <- cbind(bias.3.mat[,(dim.eval3/3+1):(dim.eval3*2/3)]/100,rmse.3.mat[,(dim.eval3/3+1):(dim.eval3*2/3)]/1000)
alpha3.results <- cbind(bias.3.mat[,(dim.eval3*2/3+1):dim.eval3]/100,rmse.3.mat[,(dim.eval3*2/3+1):dim.eval3]/1000)
         
Sys.time()-start

### Computation time

start <- Sys.time()

nrepst <- 20
nvec <- c(1000,5000,10000)

time.3.mat <- matrix(NA,length(nvec),2)

for (j in 1:length(nvec)){
  
  nr <- nvec[j]
  print(nr)
  
  
  results.t3 <- matrix(NA,nrepst,2)
  for (r in 1:nrepst){
    results.t3[r,] <- sim3.ct(x,beta50,gammaparams,minuhat,n,nr,0.5,box)
  }
  
  time.3.mat[j,] <- apply(results.t3,2,mean,na.rm=TRUE)
  
}

Sys.time()-start

#####################################################################################
# Results
####################################################################################

# Save file

filename <- paste("RData/results_bias_rmse/res_401k_3endog",format(Sys.time(),"%Y-%m-%d_%T"),".RData",sep="")
save.image(file=filename)

# Tables

print(xtable(cbind(taus,alpha1.results)),include.rownames=F)
print(xtable(cbind(taus,alpha2.results)),include.rownames=F)
print(xtable(cbind(taus,alpha3.results)),include.rownames=F)

print(xtable(cbind(nvec,time.3.mat),digits=c(0,0,rep(2,2))),include.rownames=F)




