#####################################################################################
# Simulations bias and RMSE, 1 and 2 endogenous variables
# DPG: 401(k)-Application in Chernozhukov and Hansen (2004)
# Authors: H. Kaido and K. Wuthrich
# Data source: Chernozhukov and Hansen (2004) via Chris Hansen
#####################################################################################

#####################################################################################
# Packages and preliminaries
#####################################################################################

rm(list = ls()) 

library(AER)
library(xtable)
library(fitdistrplus)
library(MASS)
library(doParallel)
library(doRNG)
registerDoParallel(cores=4)

# To replicate the results, change the working directory here.
setwd("/Users/kasparwuthrich/Dropbox/research/Decentralization IVQR/code/final_replication_package")

source("algorithms_final.R")

#####################################################################################
# Calibrate application-based simulations
#####################################################################################

# Load data

restatw     <- read.table("restatw.dat",header=T)
data.restat <- restatw[restatw$inc>=0,]
attach(data.restat)

n   <- nrow(data.restat)
y   <- net_tfa
d   <- p401
dp  <- d+1
z   <- e401
inc <- inc/10000
x   <- cbind(inc,age)
k   <- dim(x)[2]+1

# Estimate the distribution of residuals

iqr50       <- iqr(y,d,z,x,0.5,seq(0,10000,length=500)) # IQR
beta50      <- iqr50$beta 
alpha50     <- iqr50$alpha
uhat        <- y-cbind(1,x)%*%beta50-d*alpha50
minuhat     <- min(uhat)
gammaparams <- fitdist(as.vector(uhat-minuhat+1), "gamma", method="mme") 

#####################################################################################
# Simulation wrapper functions
#####################################################################################

sim1 <- function(x,beta50,gammaparams,minuhat,n,nr,tau,box){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  dr    <- zr*(0.6*temp<ur) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  drp   <- dr+1
  yr    <- cbind(1,xr)%*%beta50+dr*(5000+ur*10000)+minuhat-1+qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls        <- ivreg.fit(cbind(drp,cbind(1,xr)),yr,cbind(zr,cbind(1,xr)))  
  start.alpha <- tsls$coefficients[1]
  lower       <- start.alpha - box
  upper       <- start.alpha + box 
  leng        <- 500
  
  results <- evaluate1(yr,drp,zr,xr,tau,start.alpha,lower,upper,leng)
  
  return(results)
  
}

sim2 <- function(x,beta50,gammaparams,minuhat,n,nr,tau,box){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  d1r   <- zr*(0.6*temp<ur) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  d1rp  <- d1r+1
  z1r   <- zr
  z2r   <- rnorm(nr)
  z2rp  <- z2r-min(z2r)+1
  d2r   <- 0.8*z2r + 0.2*qnorm(ur)
  d2rp  <- d2r-min(d2r)+1
  
  yr <- cbind(1,xr)%*%beta50+d1r*(5000+ur*10000)+d2r*10000+minuhat-1+qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls <- ivreg.fit(cbind(d1rp,d2rp,cbind(1,xr)),yr,cbind(z1r,z2r,cbind(1,xr)))  
  
  start.alpha1 <- tsls$coefficients[1]
  start.alpha2 <- tsls$coefficients[2]

  lower.alpha1 <- start.alpha1-box
  upper.alpha1 <- start.alpha1+box
  lower.alpha2 <- start.alpha2-box
  upper.alpha2 <- start.alpha2+box
  leng <- 40
  
  results <- evaluate2(yr,d1rp,d2rp,z1r,z2rp,xr,tau,start.alpha1,start.alpha2,lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,leng)

  return(results)

}

#####################################################################################
# Simulations
#####################################################################################  
  
### Specifications
  
set.seed(123) 
nreps <- 500
taus <- c(0.15,0.25,0.5,0.75,0.85)
nr <- n
box <- 30000

time <- Sys.time()

### Simulation one endogenous variable

dim.eval1 <- 5

bias.1.mat <- var.1.mat <- rmse.1.mat <- convergence.1.mat <- matrix(NA,length(taus),dim.eval1)

for (j in 1:length(taus)){

  tau <- taus[j]
  print(tau)

  results.1.par <- foreach(i=1:nreps) %dorng% {
    results.1.par <- sim1(x,beta50,gammaparams,minuhat,n,nr,tau,box)
  }

  alpha10 <- (5000+tau*10000)

  results.1     <- t(matrix(unlist(results.1.par),dim.eval1,nreps))
  bias.1        <- apply(results.1,2,mean,na.rm=TRUE) - c(rep(alpha10,dim.eval1))
  var.1         <- apply(results.1,2,var,na.rm=TRUE)
  rmse.1        <- sqrt(bias.1^2 + var.1)
  convergence.1 <- colMeans(apply(results.1,2,is.na))

  bias.1.mat[j,]        <- bias.1
  var.1.mat[j,]         <- var.1
  rmse.1.mat[j,]        <- rmse.1
  convergence.1.mat[j,] <- convergence.1

}

### Simulation two endogenous variables

dim.eval2 <- 10

bias.2.mat <- var.2.mat <- rmse.2.mat <- convergence.2.mat <- matrix(NA,length(taus),dim.eval2)

for (j in 1:length(taus)){
  
  tau <- taus[j]
  print(tau)
  
  results.2.par <- foreach(i=1:nreps) %dorng% {
    results.2.par <- sim2(x,beta50,gammaparams,minuhat,n,nr,tau,box)
  }
  
  alpha10 <- (5000+tau*10000)  
  alpha20 <- 10000
  
  results.2     <- t(matrix(unlist(results.2.par),dim.eval2,nreps))
  bias.2        <- apply(results.2,2,mean,na.rm=TRUE)- c(rep(alpha10,dim.eval2/2),rep(alpha20,dim.eval2/2))
  var.2         <- apply(results.2,2,var,na.rm=TRUE)
  rmse.2        <- sqrt(bias.2^2 + var.2)
  convergence.2 <- colMeans(apply(results.2,2,is.na))
  
  bias.2.mat[j,]        <- bias.2
  var.2.mat[j,]         <- var.2
  rmse.2.mat[j,]        <- rmse.2
  convergence.2.mat[j,] <- convergence.2
  
}  

alpha1.results.2 <- cbind(bias.2.mat[,(1:(dim.eval2/2))]/100,rmse.2.mat[,(1:(dim.eval2/2))]/1000)
alpha2.results.2 <- cbind(bias.2.mat[,((dim.eval2/2+1):dim.eval2)]/100,rmse.2.mat[,((dim.eval2/2+1):dim.eval2)]/1000)

Sys.time()-time

### Save file
filename <- paste("RData/results_bias_rmse/res_401k",format(Sys.time(),"%Y-%m-%d_%T"),".RData",sep="")
save.image(file=filename)

#####################################################################################
# Results
#####################################################################################  

# Print results main text
print(xtable(cbind(taus,bias.1.mat[,-4]/100,rmse.1.mat[,-4]/1000)),include.rownames=F)

print(xtable(cbind(taus,alpha1.results.2)),include.rownames=F)
print(xtable(cbind(taus,alpha2.results.2)),include.rownames=F)

# Print comparison profiling algorithm based on Brent and Bisection (not reported in the paper)
print(xtable(cbind(taus,bias.1.mat[,3:4]/100,rmse.1.mat[,3:4]/1000)),include.rownames=F)

