#####################################################################################
# Simulations computation time
# DPG: 401(k)-Application in Chernozhukov and Hansen (2004)
# Authors: H. Kaido and K. Wuthrich
# Data source: Chernozhukov and Hansen (2004) via Chris Hansen
#####################################################################################

#####################################################################################
# Packages and preliminaries
#####################################################################################

rm(list = ls())

library(AER)
library(xtable)
library(fitdistrplus)
library(MASS)

# To replicate the results, change the working directory here.
setwd("/Users/kasparwuthrich/Dropbox/research/Decentralization IVQR/code/final_replication_package")

source("algorithms_final.R")

#####################################################################################
# Calibrate application-based simulations
#####################################################################################

# Load data

restatw   <- read.table("restatw.dat",header=T)
data.restat      <- restatw[restatw$inc>=0,]
attach(data.restat)

n   <- nrow(data.restat);
y   <- net_tfa
d   <- p401
dp  <- d+1
z   <- e401
inc <- inc/10000
x   <- cbind(inc,age)
k   <- dim(x)[2]+1

# Estimate the distribution of residuals

iqr50       <- iqr(y,d,z,x,0.5,seq(0,10000,length=500)) # IQR
beta50      <- iqr50$beta 
alpha50     <- iqr50$alpha
uhat        <- y-cbind(1,x)%*%beta50-d*alpha50
minuhat     <- min(uhat)
gammaparams <- fitdist(as.vector(uhat-minuhat+1), "gamma", method="mme") 

#####################################################################################
# Simulation wrapper functions
#####################################################################################

sim1.ct <- function(x,beta50,gammaparams,minuhat,n,nr,tau,box,leng){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  dr    <- zr*(0.6*temp<ur) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  drp   <- dr+1
  yr    <- cbind(1,xr)%*%beta50+dr*(5000+ur*10000)+minuhat-1+qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls        <- ivreg.fit(cbind(drp,cbind(1,xr)),yr,cbind(zr,cbind(1,xr)))  
  start.alpha <- tsls$coefficients[1]
  lower       <- start.alpha - box
  upper       <- start.alpha + box 

  t.br.seq          <- system.time(br.seq(yr,drp,zr,xr,tau,start.alpha))[3]
  t.brent           <- system.time(brent.uniroot(lower,upper,yr,drp,zr,xr,tau))[3]
  t.iqr             <- system.time(iqr(yr,drp,zr,xr,tau,seq(lower,upper,length=leng)))[3]
  t.brent.obj       <- system.time(brent.obj(lower,upper,yr,drp,zr,xr,tau))[3]
  t.bisection.obj   <- system.time(bisection.obj(lower,upper,yr,drp,zr,xr,tau))[3]
  
  results <- c(t.br.seq,t.brent,t.brent.obj,t.bisection.obj,t.iqr)
  
  return(results)
  
}

sim2.ct <- function(x,beta50,gammaparams,minuhat,n,nr,tau,box,leng){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  d1r   <- zr*(0.6*temp<ur) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  d1rp  <- d1r+1
  z1r   <- zr
  z2r   <- rnorm(nr)
  z2rp  <- z2r-min(z2r)+1
  d2r   <- 0.8*z2r + 0.2*qnorm(ur)
  d2rp  <- d2r-min(d2r)+1
  
  yr <- cbind(1,xr)%*%beta50+d1r*(5000+ur*10000)+d2r*10000+minuhat-1+qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls <- ivreg.fit(cbind(d1rp,d2rp,cbind(1,xr)),yr,cbind(z1r,z2r,cbind(1,xr)))  
  
  start.alpha1 <- tsls$coefficients[1]
  start.alpha2 <- tsls$coefficients[2]
  
  lower.alpha1 <- start.alpha1-box
  upper.alpha1 <- start.alpha1+box
  lower.alpha2 <- start.alpha2-box
  upper.alpha2 <- start.alpha2+box
  
  t.br.seq    <- system.time(br.seq.2(yr,d1rp,d2rp,z1r,z2rp,xr,tau,start.alpha1,start.alpha2))[3]
  t.sa        <- system.time(sa.2(yr,d1rp,d2rp,z1r,z2rp,xr,tau,start.alpha1,start.alpha2,c(lower.alpha1,lower.alpha2),c(upper.alpha1,upper.alpha2)))[3]
  t.brent     <- system.time(brent.uniroot.2(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,yr,d1rp,d2rp,z1r,z2rp,xr, tau))[3]
  t.brent.obj <- system.time(brent.obj.2(lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,yr,d1rp,d2rp,z1r,z2rp,xr, tau))[3]
  t.iqr       <- system.time(try(iqr.2(yr,d1rp,d2rp,z1r,z2rp,xr,tau,seq(lower.alpha1,upper.alpha1,length=leng),seq(lower.alpha2,upper.alpha2,length=leng)),TRUE))[3]
  
  results <- c(t.br.seq,t.brent,t.sa,t.brent.obj,t.iqr)
  
  return(results)
  
}

#####################################################################################
# Simulations
#####################################################################################

### Specifications

set.seed(123) 
box <- 30000
nrepst <- 100
nvec <- c(1000,5000,10000)

### Simulations

time.1.mat <- matrix(NA,length(nvec),5)

for (j in 1:length(nvec)){

  nr <- nvec[j]
  print(nr)
  
  results.t1 <- matrix(NA,nrepst,5)
  for (r in 1:nrepst){
    results.t1[r,] <- sim1.ct(x,beta50,gammaparams,minuhat,n,nr,0.5,box,500)
  }

  time.1.mat[j,] <- apply(results.t1,2,mean,na.rm=TRUE)

}

time.2.mat <- matrix(NA,length(nvec),5)

for (j in 1:length(nvec)){

  nr <- nvec[j]
  print(nr)
  
  results.t2 <- matrix(NA,nrepst,5)
  for (r in 1:nrepst){
    results.t2[r,] <- sim2.ct(x,beta50,gammaparams,minuhat,n,nr,0.5,box,100)
  }
  
  time.2.mat[j,] <- apply(results.t2,2,mean,na.rm=TRUE)

}

#####################################################################################
# Results
#####################################################################################

### Save file
filename <- paste("RData/results_ct/res_401k_ct",format(Sys.time(),"%Y-%m-%d_%T"),".RData",sep="")
save.image(file=filename)

### Print results main text
print(xtable(cbind(nvec,time.1.mat[,-4]),digits=c(0,0,rep(2,4))),include.rownames=F)
print(xtable(cbind(nvec,time.2.mat),digits=c(0,0,rep(2,5))),include.rownames=F)

round(time.1.mat[,5]/time.1.mat[,1])
round(time.1.mat[,5]/time.1.mat[,2])

round(time.2.mat[,5]/time.2.mat[,1])
round(time.2.mat[,5]/time.2.mat[,2])

### Comparison profiling algorithm based on Brent and Bisection (not reported in the paper)
print(xtable(cbind(nvec,time.1.mat[,3:4]),digits=c(0,0,rep(2,2))),include.rownames=F)


