#####################################################################################
# Simulation Inference
# DPG: 401(k)-Application in Chernozhukov and Hansen (2004)
# Authors: H. Kaido and K. Wuthrich
# Data source: Chernozhukov and Hansen (2004) via Chris Hansen
# This code was run on the Boston University cluster
#####################################################################################

#####################################################################################
# Packages and preliminaries
#####################################################################################

rm(list = ls())

library(AER)
library(xtable)
library(fitdistrplus)
library(MASS)
library(doParallel)
library(doRNG)
argv <-commandArgs(TRUE)
nCores <- as.numeric(argv[1])
registerDoParallel(cores=nCores)

# To replicate the results, change the working directory here.
setwd("/project/seteff/Decentralization/code/final_replication_package")

source("algorithms_final.R")

#####################################################################################
# Calibrate application-based simulations
#####################################################################################

data.original <- read.table("restatw.dat",header=T)
data          <- data.original[ data.original$inc>=0,]

n   <- nrow(data);
y   <- data$net_tfa
d   <- data$p401
dp  <- d+1
z   <- data$e401
data$inc <- data$inc/10000
x   <- cbind(data$inc,data$age)

# Estimate the distribution of residuals

iqr50       <- iqr(y,d,z,x,0.5,seq(0,10000,length=100)) # IQR
beta50      <- iqr50$beta 
alpha50     <- iqr50$alpha
uhat        <- y - cbind(1,x) %*% beta50 - d * alpha50
minuhat     <- min(uhat)
gammaparams <- fitdist(as.vector(uhat-minuhat+1),"gamma",method="mme") 

#####################################################################################
# Simulation wrapper function
#####################################################################################

sim1.inf <- function(x,beta50,gammaparams,minuhat,nr,tau,alpha.sig,bsreps,box){
  
  xr    <- x[sample(n,nr,replace=T), ]
  ur    <- runif(nr)
  zr    <- rbinom(nr, 1, mean(z)) 
  temp  <- runif(nr)
  dr    <- 1*(zr==1&(0.6*temp<ur)) #chosen to roughly match distribution of (d,z): cf. mean(dr) mean(dr[zr==1]) mean(d) mean(d[z==1])
  drp   <- dr+1
  
  yr <- cbind(1,xr) %*% beta50 + dr*(5000+ur*10000) + minuhat - 1 + qgamma(ur,shape=gammaparams$estimate[1], rate=gammaparams$estimate[2])
  
  tsls        <- ivreg.fit(cbind(drp,cbind(1,xr)),yr,cbind(zr,cbind(1,xr)))  
  start.alpha <- tsls$coefficients[1]
  lower       <- start.alpha - box
  upper       <- start.alpha + box

  br.seq.alpha  <- br.seq(yr,drp,zr,xr,tau,start.alpha)$alpha
  brent.alpha   <- brent.uniroot(lower,upper,yr,drp,zr,xr,tau)$alpha
  
  brent.bs  <- matrix(NA,bsreps,1) 
  br.seq.bs <- matrix(NA,bsreps,1) 
  
  for (rep in 1:bsreps){
    
    indexs <- sample(nr,nr,replace=TRUE)
    ys      <- yr[indexs]
    ds      <- dr[indexs]
    dsp     <- ds+1
    zs      <- zr[indexs]
    xs      <- xr[indexs,]
    
    tsls.bs        <- ivreg.fit(cbind(dsp,cbind(1,xs)),ys,cbind(zs,cbind(1,xs)))  
    start.alpha.bs <- tsls.bs$coefficients[1]
    
    br.seq.bs[rep,1]  <- br.seq(ys,dsp,zs,xs,tau,start.alpha.bs)$alpha
    brent.bs[rep,1]   <- brent.uniroot(start.alpha.bs-box,start.alpha.bs+box,ys,dsp,zs,xs,tau)$alpha 

  }
  
  # SE estimation
  sd.brent    <- sqrt(mean((brent.bs-brent.alpha)^2,na.rm=T))
  ci.brent.u  <- brent.alpha + qnorm(1-alpha.sig/2) * sd.brent
  ci.brent.l  <- brent.alpha - qnorm(1-alpha.sig/2) * sd.brent
  
  sd.br.seq    <- sqrt(mean((br.seq.bs-br.seq.alpha)^2,na.rm=T))
  ci.br.seq.u  <- br.seq.alpha + qnorm(1-alpha.sig/2) * sd.br.seq
  ci.br.seq.l  <- br.seq.alpha - qnorm(1-alpha.sig/2) * sd.br.seq

  alpha01     <- 5000 + tau * 10000
  cov.br.seq  <- (alpha01>=ci.br.seq.l)*(alpha01<=ci.br.seq.u)
  cov.brent  <- (alpha01>=ci.brent.l)*(alpha01<=ci.brent.u)
  
  cov <- c(cov.br.seq,cov.brent)
  sd  <- c(sd.brent,sd.br.seq)
  ci.brent <- c(ci.brent.l,ci.brent.u)
  ci.br.seq <- c(ci.br.seq.l,ci.br.seq.u)
  results <- list("cov"=cov,"sd"=sd,"ci.brent"=ci.brent,"ci.br.seq"=ci.br.seq)
  return(results)
  
}

#####################################################################################
# Simulations
#####################################################################################  

### Setting
set.seed(123) 
nreps       <- 1000
bsreps      <- 1000
alpha.sigs  <- c(0.1,0.05)
taus        <- c(0.15,0.25,0.5,0.75,0.85)
box         <- 30000
# coverage.array <- array(NA,c(length(taus),2,length(alpha.sigs)))

### Simulation loop
ind <- as.integer(argv[2]) # this number is between 1 & 10
if(ind<=5) alpha.sig <- 0.1 else alpha.sig <- 0.05
ind2 <- ind%%5 
if(ind2==0) ind2 <- 5 
tau <- taus[ind2] 

print(c(tau,alpha.sig))
system.time(
results.par <- foreach(i=1:nreps) %dorng% {
  results.par <- sim1.inf(x,beta50,gammaparams,minuhat,n,tau,alpha.sig,bsreps,box)
})

results <- list()
results$cov <- t(sapply(results.par,function(x){return (x$cov)}))
results$sd <- t(sapply(results.par,function(x){return (x$sd)}))
results$ci.brent  <- t(sapply(results.par,function(x){return (x$ci.brent)}))
results$ci.br.seq <- t(sapply(results.par,function(x){return (x$ci.br.seq)}))
results$length.brent <- results$ci.brent[,2]-results$ci.brent[,1]
results$length.br.seq <- results$ci.br.seq[,2]-results$ci.br.seq[,1]
coverage  <- apply(results$cov,2,mean,na.rm=TRUE)
mean.sd   <- apply(results$sd,2,mean,na.rm=TRUE)
length.brent <- mean(results$length.brent,na.rm=TRUE)
length.br.seq <- mean(results$length.br.seq,na.rm=TRUE)
stopImplicitCluster()

### Results

filename <- paste("RData/results_inference/Res_tau",tau,"alpha",alpha.sig,format(Sys.time(),"%Y-%m-%d_%T"),".RData",sep="")
save.image(file=filename)


