#####################################################################################
# Simulation
# DPG: Location Scale (Andrews and Mikusheva, 2016)
# Authors: H. Kaido and K. Wuthrich
#####################################################################################

#####################################################################################
# Packages and preliminaries
#####################################################################################

rm(list = ls())

library(AER)
library(xtable)
library(fitdistrplus)
library(MASS)
library(doParallel)
library(doRNG)
registerDoParallel(cores=4)

# To replicate the results, change the working directory here.
setwd("/Users/kasparwuthrich/Dropbox/research/Decentralization IVQR/code/replication_package_final")

source("algorithms_final.R")

#####################################################################################
# Simulation wrapper functions
#####################################################################################

sim1sy <- function(n,tau,box){
  
  gamma1 <- gamma2 <- gamma3 <- gamma4 <- gamma5 <- 1
  
  r1 <- 0.5
  p1 <- 0.8
  
  mu <- c(0,0,0)
  
  Sigma <- rbind(c(1,r1,0),c(r1,1,p1),c(0,p1,1))
  
  epsilon  <- mvrnorm(n,mu,Sigma)
  epsilon1 <- epsilon[,1]
  epsilon2 <- epsilon[,2]
  epsilon3 <- epsilon[,3]
  
  u <- pnorm(epsilon1)#-(0.5)
  d <- pnorm(epsilon2)
  z <- pnorm(epsilon3)
  x <- pnorm(rnorm(n))
  
  y <- gamma1 + gamma2*x + gamma3*d + (gamma4 + gamma5*d)*u 
  
  tsls <- ivreg.fit(cbind(d,cbind(1,x)),y,cbind(z,cbind(1,x)))  
  
  start.alpha <- tsls$coefficients[1]
  
  lower <- start.alpha-box
  upper <- start.alpha+box
  
  leng <- 500
  
  results <- evaluate1(y,d,z,x,tau,start.alpha,lower,upper,leng)
  
  return(results)
  
}

sim1asy <- function(n,tau,box){
  
  gamma1 <- gamma2 <- gamma3 <- gamma4 <- gamma5 <- 1
  
  r1 <- 0.5
  p1 <- 0.8
  
  mu <- c(0,0,0)
  
  Sigma <- rbind(c(1,r1,0),c(r1,1,p1),c(0,p1,1))
  
  epsilon  <- mvrnorm(n,mu,Sigma)
  epsilon1 <- epsilon[,1]
  epsilon2 <- epsilon[,2]
  epsilon3 <- epsilon[,3]
  
  u <- epsilon1
  d <- exp(2*epsilon2)
  z <- epsilon3-min(epsilon3)+0.1
  x <- rnorm(n)
  
  y <- gamma1 + gamma2*x + gamma3*d + (gamma4 + gamma5*d)*u 
  
  tsls <- ivreg.fit(cbind(d,cbind(1,x)),y,cbind(z,cbind(1,x)))  
  
  start.alpha <- tsls$coefficients[1]
  
  lower <- start.alpha-box
  upper <- start.alpha+box
  
  leng <- 500
  
  results <- evaluate1(y,d,z,x,tau,start.alpha,lower,upper,leng)
  
  return(results)
  
}

sim2sy <- function(n,tau,box){
  
  gamma1 <- gamma2 <- gamma3 <- gamma4 <- gamma5 <- gamma6 <- gamma7 <- 1
  
  r1  <- 0.5
  r2  <- 0.5
  p11 <- 0.8
  p21 <- 0
  p12 <- 0
  p22 <- 0.4
  pz  <- 0
  
  mu <- c(0,0,0,0,0)
  
  Sigma <- rbind(c(1,r1,r2,0,0),c(r1,1,0,p11,p12),c(r2,0,1,p21,p22),c(0,p11,p21,1,pz),c(0,p12,p22,pz,1))
  
  epsilon  <- mvrnorm(n,mu,Sigma)
  epsilon1 <- epsilon[,1]
  epsilon2 <- epsilon[,2]
  epsilon3 <- epsilon[,3]
  epsilon4 <- epsilon[,4]
  epsilon5 <- epsilon[,5]
  
  u  <- pnorm(epsilon1)#-(0.5)
  d1 <- pnorm(epsilon2)
  d2 <- pnorm(epsilon3)
  z1 <- pnorm(epsilon4)
  z2 <- pnorm(epsilon5)
  x  <- pnorm(rnorm(n))
  
  y <- gamma1 + gamma2*x + gamma3*d1 + gamma4*d2 + (gamma5 + gamma6*d1 + gamma7*d2)*u 
  
  tsls <- ivreg.fit(cbind(d1,d2,cbind(1,x)),y,cbind(z1,z2,cbind(1,x)))  

  start.alpha1 <- tsls$coefficients[1]
  start.alpha2 <- tsls$coefficients[2]
  lower.alpha1 <- start.alpha1-box
  upper.alpha1 <- start.alpha1+box
  lower.alpha2 <- start.alpha2-box
  upper.alpha2 <- start.alpha2+box

  leng <- 40
  
  results <- evaluate2(y,d1,d2,z1,z2,x,tau,start.alpha1,start.alpha2,lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,leng)
  
  return(results)
 
}

sim2asy = function(n,tau,box){
  
  gamma1 <- gamma2 <- gamma3 <- gamma4 <- gamma5 <- gamma6 <- gamma7 <- 1
  
  r1  <- 0.5
  r2  <- 0.5
  p11 <- 0.8
  p21 <- 0
  p12 <- 0
  p22 <- 0.4
  pz  <- 0
  
  mu <- c(0,0,0,0,0)
  
  Sigma <- rbind(c(1,r1,r2,0,0),c(r1,1,0,p11,p12),c(r2,0,1,p21,p22),c(0,p11,p21,1,pz),c(0,p12,p22,pz,1))
  
  epsilon  <- mvrnorm(n,mu,Sigma)
  epsilon1 <- epsilon[,1]
  epsilon2 <- epsilon[,2]
  epsilon3 <- epsilon[,3]
  epsilon4 <- epsilon[,4]
  epsilon5 <- epsilon[,5]
  
  u  <- epsilon1
  d1 <- exp(2*epsilon2)
  d2 <- exp(2*epsilon3)
  z1 <- epsilon4-min(epsilon4)+0.1
  z2 <- epsilon5-min(epsilon5)+0.1
  x  <- rnorm(n)
  
  y <- gamma1 + gamma2*x + gamma3*d1 + gamma4*d2 + (gamma5 + gamma6*d1 + gamma7*d2)*u 
  
  tsls <- ivreg.fit(cbind(d1,d2,cbind(1,x)),y,cbind(z1,z2,cbind(1,x)))  
  
  start.alpha1 <- tsls$coefficients[1]
  start.alpha2 <- tsls$coefficients[2]
  lower.alpha1 <- start.alpha1-box
  upper.alpha1 <- start.alpha1+box
  lower.alpha2 <- start.alpha2-box
  upper.alpha2 <- start.alpha2+box
  leng <- 40
  
  results <- evaluate2(y,d1,d2,z1,z2,x,tau,start.alpha1,start.alpha2,lower.alpha1,upper.alpha1,lower.alpha2,upper.alpha2,leng)

  return(results)
  
}

#################################################################################################
# Simulations
#################################################################################################

### Specification

set.seed(123) 
nreps <- 500
ns <- c(500,1000)
taus <- c(0.15,0.25,0.5,0.75,0.85)
box <- 4

### Simulations one endogenous variable

dim.eval1 <- 5

bias.array.1.sy <- var.array.1.sy <- rmse.array.1.sy <- bias.array.1.asy <- var.array.1.asy <- rmse.array.1.asy <- convergence.1.sy <- convergence.1.asy <- array(NA,c(length(taus),dim.eval1,length(ns)))

for (i in 1:length(ns)){
  for (j in 1:length(taus)){

    n <- ns[i]
    tau <- taus[j]

    print(c(tau,n))

    results.par.sy <- foreach(i=1:nreps) %dorng% {
      results.par.sy <- sim1sy(n,tau,box)
    }

    results.sy <- t(matrix(unlist(results.par.sy),dim.eval1,nreps))
    bias.sy <- apply(results.sy,2,mean,na.rm=TRUE)-c(rep(1+tau,dim.eval1))
    var.sy <- apply(results.sy,2,var,na.rm=TRUE)
    rmse.sy <- sqrt(bias.sy^2 + var.sy)
    convergence.sy <- colMeans(apply(results.sy,2,is.na))

    bias.array.1.sy[j,,i] <- bias.sy
    var.array.1.sy[j,,i] <- var.sy
    rmse.array.1.sy[j,,i] <- rmse.sy
    convergence.1.sy[j,,i] <- convergence.sy

    results.par.asy <- foreach(i=1:nreps) %dorng% {
      results.par.asy <- sim1asy(n,tau,box)
    }

    results.asy <- t(matrix(unlist(results.par.asy),dim.eval1,nreps))
    bias.asy <- apply(results.asy,2,mean,na.rm=TRUE)-c(rep(1+qnorm(tau),dim.eval1))
    var.asy <- apply(results.asy,2,var,na.rm=TRUE)
    rmse.asy <- sqrt(bias.asy^2 + var.asy)
    convergence.asy <- colMeans(apply(results.asy,2,is.na))
    
    bias.array.1.asy[j,,i] <- bias.asy
    var.array.1.asy[j,,i] <- var.asy
    rmse.array.1.asy[j,,i] <- rmse.asy
    convergence.1.asy[j,,i] <- convergence.asy

  }
}

# Symmetric Design

bias.1.sy.n1 <- bias.array.1.sy[,,1]
bias.1.sy.n2 <- bias.array.1.sy[,,2]

rmse.1.sy.n1 <- rmse.array.1.sy[,,1]
rmse.1.sy.n2 <- rmse.array.1.sy[,,2]

results.1.sy.alpha1.n1 <- cbind(bias.1.sy.n1,rmse.1.sy.n1)
results.1.sy.alpha1.n2 <- cbind(bias.1.sy.n2,rmse.1.sy.n2)

# Asymmetric Design

bias.1.asy.n1 <- bias.array.1.asy[,,1]
bias.1.asy.n2 <- bias.array.1.asy[,,2]

rmse.1.asy.n1 <- rmse.array.1.asy[,,1]
rmse.1.asy.n2 <- rmse.array.1.asy[,,2]

results.1.asy.alpha1.n1 <- cbind(bias.1.asy.n1,rmse.1.asy.n1)
results.1.asy.alpha1.n2 <- cbind(bias.1.asy.n2,rmse.1.asy.n2)

### Simulations two endogenous variables

dim.eval2 <- 10

bias.array.2.sy <- var.array.2.sy <- rmse.array.2.sy <- bias.array.2.asy <- var.array.2.asy <- rmse.array.2.asy <- convergence.2.sy <- convergence.2.asy <- array(NA,c(length(taus),dim.eval2,length(ns)))

for (i in 1:length(ns)){
for (j in 1:length(taus)){
  
  n <- ns[i]
  tau <- taus[j]
  
  print(c(tau,n))
  
  results.par.sy <- foreach(i=1:nreps) %dorng% {
    results.par.sy <- sim2sy(n,tau,box)
  }

  results.sy <- t(matrix(unlist(results.par.sy),dim.eval2,nreps))
  bias.sy <- apply(results.sy,2,mean,na.rm=TRUE)-c(rep(1+tau,(dim.eval2/2)),rep(1+tau,(dim.eval2/2)))
  var.sy <- apply(results.sy,2,var,na.rm=TRUE)
  rmse.sy <- sqrt(bias.sy^2 + var.sy)
  convergence.sy <- colMeans(apply(results.sy,2,is.na))
  
  bias.array.2.sy[j,,i] <- bias.sy
  var.array.2.sy[j,,i] <- var.sy
  rmse.array.2.sy[j,,i] <- rmse.sy
  convergence.2.sy[j,,i] <- convergence.sy
  
  results.par.asy <- foreach(i=1:nreps) %dorng% {
    results.par.asy <- sim2asy(n,tau,box)
  }
  
  results.asy <- t(matrix(unlist(results.par.asy),dim.eval2,nreps))
  bias.asy <- apply(results.asy,2,mean,na.rm=TRUE)-c(rep(1+qnorm(tau),(dim.eval2/2)),rep(1+qnorm(tau),(dim.eval2/2)))
  var.asy <- apply(results.asy,2,var,na.rm=TRUE)
  rmse.asy <- sqrt(bias.asy^2 + var.asy)
  convergence.asy <- colMeans(apply(results.asy,2,is.na))
  
  bias.array.2.asy[j,,i] <- bias.asy
  var.array.2.asy[j,,i] <- var.asy
  rmse.array.2.asy[j,,i] <- rmse.asy
  convergence.2.asy[j,,i] <- convergence.asy
  
}  
}

# Symmetric Design

bias.2.sy.n1 <- bias.array.2.sy[,,1]
bias.2.sy.n2 <- bias.array.2.sy[,,2]

rmse.2.sy.n1 <- rmse.array.2.sy[,,1]
rmse.2.sy.n2 <- rmse.array.2.sy[,,2]

results.2.sy.alpha1.n1 <- cbind(bias.2.sy.n1[,(1:(dim.eval2/2))],rmse.2.sy.n1[,(1:(dim.eval2/2))])
results.2.sy.alpha2.n1 <- cbind(bias.2.sy.n1[,((dim.eval2/2+1):dim.eval2)],rmse.2.sy.n1[,((dim.eval2/2+1):dim.eval2)])

results.2.sy.alpha1.n2 <- cbind(bias.2.sy.n2[,(1:(dim.eval2/2))],rmse.2.sy.n2[,(1:(dim.eval2/2))])
results.2.sy.alpha2.n2 <- cbind(bias.2.sy.n2[,((dim.eval2/2+1):dim.eval2)],rmse.2.sy.n2[,((dim.eval2/2+1):dim.eval2)])

# Asymmetric Design

bias.2.asy.n1 <- bias.array.2.asy[,,1]
bias.2.asy.n2 <- bias.array.2.asy[,,2]

rmse.2.asy.n1 <- rmse.array.2.asy[,,1]
rmse.2.asy.n2 <- rmse.array.2.asy[,,2]

results.2.asy.alpha1.n1 <- cbind(bias.2.asy.n1[,(1:(dim.eval2/2))],rmse.2.asy.n1[,(1:(dim.eval2/2))])
results.2.asy.alpha2.n1 <- cbind(bias.2.asy.n1[,((dim.eval2/2+1):dim.eval2)],rmse.2.asy.n1[,((dim.eval2/2+1):dim.eval2)])

results.2.asy.alpha1.n2 <- cbind(bias.2.asy.n2[,(1:(dim.eval2/2))],rmse.2.asy.n2[,(1:(dim.eval2/2))])
results.2.asy.alpha2.n2 <- cbind(bias.2.asy.n2[,((dim.eval2/2+1):dim.eval2)],rmse.2.asy.n2[,((dim.eval2/2+1):dim.eval2)])

### Save file

filename <- paste("RData/results_bias_rmse/res_locscale",format(Sys.time(),"%Y-%m-%d_%T"),".RData",sep="")
save.image(file=filename)


#################################################################################################
# Tables
#################################################################################################

# One endogenous variable (without bisection -> take out columns 4 and 9)

print(xtable(cbind(taus,results.1.sy.alpha1.n1[,-c(4,9)])),include.rownames=F) 
print(xtable(cbind(taus,results.1.sy.alpha1.n2[,-c(4,9)])),include.rownames=F)

print(xtable(cbind(taus,results.1.asy.alpha1.n1[,-c(4,9)])),include.rownames=F)
print(xtable(cbind(taus,results.1.asy.alpha1.n2[,-c(4,9)])),include.rownames=F)

# Two endogenous variables

print(xtable(cbind(taus,results.2.sy.alpha1.n1)),include.rownames=F)
print(xtable(cbind(taus,results.2.sy.alpha2.n1)),include.rownames=F)

print(xtable(cbind(taus,results.2.sy.alpha1.n2)),include.rownames=F)
print(xtable(cbind(taus,results.2.sy.alpha2.n2)),include.rownames=F)

print(xtable(cbind(taus,results.2.asy.alpha1.n1)),include.rownames=F)
print(xtable(cbind(taus,results.2.asy.alpha2.n1)),include.rownames=F)

print(xtable(cbind(taus,results.2.asy.alpha1.n2)),include.rownames=F)
print(xtable(cbind(taus,results.2.asy.alpha2.n2)),include.rownames=F)

