#####################################################################################
# Simulation Inference
# DPG: Location Scale (Andrews and Mikusheva, 2016)
# Authors: H. Kaido and K. Wuthrich
# This code was run on the Boston University cluster
#####################################################################################

#####################################################################################
# Packages and preliminaries
#####################################################################################

rm(list = ls())

library(AER)
library(xtable)
library(fitdistrplus)
library(MASS)
library(doParallel)
library(doRNG)
argv <-commandArgs(TRUE)
nCores <- as.numeric(argv[1])
registerDoParallel(cores=nCores)

# To replicate the results, change the working directory here.
setwd("/project/seteff/Decentralization/code/final_replication_package")

source("algorithms_final.R")

#####################################################################################
# Simulation wrapper function
#####################################################################################

sim1sy <- function(n,tau,bsreps,alpha.sig,box){
  
  gamma1 <- gamma2 <- gamma3 <- gamma4 <- gamma5 <- 1
  
  r1 <- 0.5
  p1 <- 0.5
  
  mu <- c(0,0,0)
  
  Sigma <- rbind(c(1,r1,0),c(r1,1,p1),c(0,p1,1))
  
  epsilon <- mvrnorm(n,mu,Sigma)
  epsilon1 <-  epsilon[,1]
  epsilon2 <-  epsilon[,2]
  epsilon3 <-  epsilon[,3]
  
  u <- pnorm(epsilon1)#-(0.5)
  d <- pnorm(epsilon2)
  z <- pnorm(epsilon3)
  x <- pnorm(rnorm(n))
  
  y <- gamma1 + gamma2*x + gamma3*d + (gamma4 + gamma5*d)*u 
  
  tsls <- ivreg.fit(cbind(d,cbind(1,x)),y,cbind(z,cbind(1,x)))  
  
  start.alpha <- tsls$coefficients[1]
  lower <- start.alpha-box
  upper <- start.alpha+box
  
  br.seq.alpha <- br.seq(y,d,z,x,tau,start.alpha)$alpha
  brent.alpha <- brent.uniroot(lower,upper,y,d,z,x,tau)$alpha
  
  brent.bs <- matrix(NA,bsreps,1) 
  br.seq.bs <- matrix(NA,bsreps,1) 
  
  for (rep in 1:bsreps){
    index.s <- sample(n,n,replace=TRUE)
    ys <- y[index.s]
    ds <- d[index.s]
    zs <- z[index.s]
    xs <- x[index.s]
    
    tsls.bs <- ivreg.fit(cbind(ds,cbind(1,xs)),ys,cbind(zs,cbind(1,xs)))  
    
    start.alpha.bs <- tsls.bs$coefficients[1]
    lower.bs <- start.alpha.bs-box
    upper.bs <- start.alpha.bs+box
    
    br.seq.bs[rep,1] <- br.seq(ys,ds,zs,xs,tau,start.alpha.bs)$alpha
    brent.bs[rep,1] <- brent.uniroot(lower.bs,upper.bs,ys,ds,zs,xs,tau)$alpha 
    
  }
  
  # different SE estimators
  
  sd.br.seq <- sqrt(mean((br.seq.bs-br.seq.alpha)^2,na.rm=T))
  sd.brent <- sqrt(mean((brent.bs-brent.alpha)^2,na.rm=T))
  
  ci.brent.u <- brent.alpha + qnorm(1-alpha.sig/2) * sd.brent
  ci.brent.l <- brent.alpha - qnorm(1-alpha.sig/2) * sd.brent
  
  ci.br.seq.u <- br.seq.alpha + qnorm(1-alpha.sig/2) * sd.br.seq
  ci.br.seq.l <- br.seq.alpha - qnorm(1-alpha.sig/2) * sd.br.seq
  
  alpha01 <- 1+tau
  cov.br.seq <- (alpha01>=ci.br.seq.l)*(alpha01<=ci.br.seq.u)
  cov.brent <- (alpha01>=ci.brent.l)*(alpha01<=ci.brent.u)
  cov <- c(cov.br.seq,cov.brent)
  sd  <- c(sd.brent,sd.br.seq)
  ci.brent <- c(ci.brent.l,ci.brent.u)
  ci.br.seq <- c(ci.br.seq.l,ci.br.seq.u)
  results <- list("cov"=cov,"sd"=sd,"ci.brent"=ci.brent,"ci.br.seq"=ci.br.seq)
  return(results)
}

sim1asy <- function(n,tau,bsreps,alpha.sig,box){
  
  gamma1 <- gamma2 <- gamma3 <- gamma4 <- gamma5 <- 1
  
  r1 <- 0.5
  p1 <- 0.5
  
  mu <- c(0,0,0)
  
  Sigma <- rbind(c(1,r1,0),c(r1,1,p1),c(0,p1,1))
  
  epsilon <- mvrnorm(n,mu,Sigma)
  epsilon1 <-  epsilon[,1]
  epsilon2 <-  epsilon[,2]
  epsilon3 <-  epsilon[,3]
  
  u <- epsilon1
  d <- exp(2*epsilon2)
  z <- epsilon3-min(epsilon3)+0.1
  x <- rnorm(n)
  
  y <- gamma1 + gamma2*x + gamma3*d + (gamma4 + gamma5*d)*u 
  
  tsls <- ivreg.fit(cbind(d,cbind(1,x)),y,cbind(z,cbind(1,x)))  
  
  start.alpha <- tsls$coefficients[1]
  lower <- start.alpha-box
  upper <- start.alpha+box
  
  br.seq.alpha <- br.seq(y,d,z,x,tau,start.alpha)$alpha
  brent.alpha <- brent.uniroot(lower,upper,y,d,z,x,tau)$alpha
  
  brent.bs <- matrix(NA,bsreps,1) 
  br.seq.bs <- matrix(NA,bsreps,1) 
  
  for (rep in 1:bsreps){
    index.s <- sample(n,n,replace=TRUE)
    ys <- y[index.s]
    ds <- d[index.s]
    zs <- z[index.s]
    xs <- x[index.s]
    
    tsls.bs <- ivreg.fit(cbind(ds,cbind(1,xs)),ys,cbind(zs,cbind(1,xs)))  
    
    start.alpha.bs <- tsls.bs$coefficients[1]
    lower.bs <- start.alpha.bs-box
    upper.bs <- start.alpha.bs+box
    
    br.seq.bs[rep,1] <- br.seq(ys,ds,zs,xs,tau,start.alpha.bs)$alpha
    brent.bs[rep,1] <- brent.uniroot(lower.bs,upper.bs,ys,ds,zs,xs,tau)$alpha 
    
  }
  
  sd.br.seq <- sqrt(mean((br.seq.bs-br.seq.alpha)^2,na.rm=T))
  sd.brent <- sqrt(mean((brent.bs-brent.alpha)^2,na.rm=T))
  
  ci.brent.u <- brent.alpha + qnorm(1-alpha.sig/2) * sd.brent
  ci.brent.l <- brent.alpha - qnorm(1-alpha.sig/2) * sd.brent
  
  ci.br.seq.u <- br.seq.alpha + qnorm(1-alpha.sig/2) * sd.br.seq
  ci.br.seq.l <- br.seq.alpha - qnorm(1-alpha.sig/2) * sd.br.seq
  
  alpha01 <- 1+qnorm(tau)
  cov.br.seq <- (alpha01>=ci.br.seq.l)*(alpha01<=ci.br.seq.u)
  cov.brent <- (alpha01>=ci.brent.l)*(alpha01<=ci.brent.u)
  cov <- c(cov.br.seq,cov.brent)
  sd  <- c(sd.brent,sd.br.seq)
  ci.brent <- c(ci.brent.l,ci.brent.u)
  ci.br.seq <- c(ci.br.seq.l,ci.br.seq.u)
  results <- list("cov"=cov,"sd"=sd,"ci.brent"=ci.brent,"ci.br.seq"=ci.br.seq)
  return(results)
}

#####################################################################################
# Simulations
#####################################################################################  

# Setting
set.seed(123) 
nreps <- 1000
bsreps <- 1000
n <- 500
taus <- c(0.15,0.25,0.5,0.75,0.85)
# alpha.sigs <- c(0.05,0.1)
ind <- as.integer(argv[2]) # this number is between 1 & 10
if(ind<=5) alpha.sig <- 0.1 else alpha.sig <- 0.05
ind2 <- ind%%5 
if(ind2==0) ind2 <- 5 
tau <- taus[ind2] 

box <- 4

# coverage.array.sy <- coverage.array.asy <- array(NA,c(length(taus),2,length(alpha.sigs)))

### Simulation
print(c(tau,alpha.sig))

### Symmetric case
results.par.sy <- foreach(i=1:nreps) %dorng% {
  results.par.sy <- sim1sy(n,tau,bsreps,alpha.sig,box)
}
results.sy<- list()
results.sy <- t(matrix(unlist(results.par.sy),2,nreps))
results.sy$cov <- t(sapply(results.par.sy,function(x){return (x$cov)}))
results.sy$sd <- t(sapply(results.par.sy,function(x){return (x$sd)}))
results.sy$ci.brent  <- t(sapply(results.par.sy,function(x){return (x$ci.brent)}))
results.sy$ci.br.seq <- t(sapply(results.par.sy,function(x){return (x$ci.br.seq)}))
results.sy$length.brent <- results.sy$ci.brent[,2]-results.sy$ci.brent[,1]
results.sy$length.br.seq <- results.sy$ci.br.seq[,2]-results.sy$ci.br.seq[,1]
coverage.sy  <- apply(results.sy$cov,2,mean,na.rm=TRUE)
mean.sd.sy   <- apply(results.sy$sd,2,mean,na.rm=TRUE)
length.brent.sy <- mean(results.sy$length.brent,na.rm=TRUE)
length.br.seq.sy <- mean(results.sy$length.br.seq,na.rm=TRUE)

### Asymmetric DGP
results.par.asy <- foreach(i=1:nreps) %dorng% {
  results.par.asy <- sim1asy(n,tau,bsreps,alpha.sig,box)
}
results.asy<- list()
results.asy <- t(matrix(unlist(results.par.asy),2,nreps))
results.asy$cov <- t(sapply(results.par.asy,function(x){return (x$cov)}))
results.asy$sd <- t(sapply(results.par.asy,function(x){return (x$sd)}))
results.asy$ci.brent  <- t(sapply(results.par.asy,function(x){return (x$ci.brent)}))
results.asy$ci.br.seq <- t(sapply(results.par.asy,function(x){return (x$ci.br.seq)}))
results.asy$length.brent <- results.asy$ci.brent[,2]-results.asy$ci.brent[,1]
results.asy$length.br.seq <- results.asy$ci.br.seq[,2]-results.asy$ci.br.seq[,1]
coverage.asy  <- apply(results.asy$cov,2,mean,na.rm=TRUE)
mean.sd.asy   <- apply(results.asy$sd,2,mean,na.rm=TRUE)
length.brent.asy <- mean(results.asy$length.brent,na.rm=TRUE)
length.br.seq.asy <- mean(results.asy$length.br.seq,na.rm=TRUE)

stopImplicitCluster()

### Results

filename <- paste("RData/results_inference/Res_locscal_tau",tau,"alpha",alpha.sig,format(Sys.time(),"%Y-%m-%d_%T"),".RData",sep="")
save.image(file=filename)
